/* 	ArgumentMap.java
 * 
 * 	Created on: Oct 2, 2002
 * 	Created by: Dennis Groenendijk
 * 
 * Copyright (c) 2002-2006, Brains2B.org
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * @#History 
 * [When]       [Who]               [What] 
 * 25-nov-2006  dennis@brains2b.nl  Fixed problem with sequence of arguments, desynchronized Map and List
 * 27-06-2007   dennis@brains2b.nl  #117 Possibility to supply sinlue character options in condensed format
 */
package org.brains2b.console;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/** An ArgumentMap is a 'road map' to the arguments that can be supplied to Main
 *  and the expected follow up.
 * <p>arguments and options are read in the order they where supplied in by <code>static void main(String[] args)</code> 
 * <p>The arguments and their key are stored in a HashMap fur future retrieval.
 * <p> You can either add options, which could be translated as switches or add arguments
 * where the entire argument which is supplied by <code>static void main(String[] args)</code> is
 * supplied to Main.
 * <p> You can add options to an argument. Which means that the value of an argument
 * will have to be one of the options, otherwise the argument will be ignored if it is optional
 * or throw an ArgumentException if it is not optional.
 * <p> You can add an argument to an option. Which means that main after finding
 * the option looks for an argument belonging to it. This argument has its own name and can 
 * be later referenced by this name.
 * <p>Options can have next to the required long form, an alternative optional one character short form. supplying options
 * to Java will either be --longform or -shortform in good Unix fashion. Both forms will signify the same option.
 * 
 *
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version $Revision: 1.5 $ $Date: 2009/09/15 08:09:58 $
 */
public class ArgumentMap {

	/** An inner class to hold the individual argument information.
	 *  <p>All options and arguments are internally represented as an ArgumentEntry
	 *  <p>All access is done to accessors.
	 */
	class ArgumentEntry {
		private String m_name;
		private char m_shortKey;
		private boolean m_required;
		private String m_parent;
		private String m_default;
		private String m_value;
		private ArrayList  m_children;
		
		/**
		 * Constructor for ArgumentEntry.
		 * <p> Initializes the complete object without children.
		 * @param name String, the long name of the argument
		 * @param shortKey char, the short form of the argument
		 * @param required boolean, true if argument should always be supplied
		 * @param parent String, the argument name of the Argument Entry for which this is a parent.
		 */
		public ArgumentEntry(String name,  char shortKey, boolean required, String parent) {
			m_name=name;
			m_shortKey=shortKey;
			m_required=required;
			m_parent=parent;
			
		}
		
		/**
		 * gets the one character short key for this argument
		 * @return char
		 */
		public char getShortKey() {
			return m_shortKey;
		}
		
		/**
		 * Adds a child to this argument entry, which can be an argument to an option
		 * or an option to an argument.
		 * <p>The parent for the supplied child will <b>not</b> automatically be set.
		 * @param name String
		 */
		public void addChild(String name) {
			if (m_children == null) {
				m_children=new ArrayList();
			}
			m_children.add(name);
		}
		
		/**
		 * gets all the Children defined for this ArgumentEntry
		 * @return List of all the children.
		 */
		public List getChildren() {
			return m_children;
		}
		
		/**
		 * get the name for this ArgumentEntry
		 * @return String, the name
		 */
		public String getName() {
			return m_name;
		}
		
		/**
		 * get the parent for this ArgumentEntry
		 * @return String, the parent
		 */
		public String getParent() {
			return m_parent;
		}
		
		/**
		 * get the default value for this ArgumentEntry
		 * <p>default values are only supplied for arguments
		 * @return String, the default value
		 */
		public String getDefaultValue() {
			return m_default;
		}
		
		/**
		 * get the value for this ArgumentEntry
		 * <p>values are only supplied for arguments, not for options
		* @return String, the value for this argument
		 */
		public String getValue() {
			return m_value;
		}
		
		/**
		 * sets the value for this ArgumentEntry
		 * <p>values are only supplied for arguments, not for options
		 * @param value String, the value
		 */
		public void setValue(String value) {
			m_value=value;
		}
		
		/**
		 * sets the default value for this ArgumentEntry
		 * <p>default values are only supplied for arguments
		 * @param value String,  the default value
		 */
		public void setDefaultValue(String value) {
			m_default=value;
			m_value=value;
		}
		
		/**
		 * checks to see if this ArgumentEntry is required
		 * @return boolean, true is ArgumentEntry is required
		 */
		public boolean isRequired() {
			return m_required;
		}
	}

	private ArrayList m_args;

	/**
	 * Constructor for ArgumentMap.
	 */
	public ArgumentMap() {
		super();
		m_args=new ArrayList();
	}
	
	
	/**
	 * adds an Argument to the root of this map. 
	 * <p>A root can have multiple arguments and options
	 * @param name String,  the name for this argument
	 * @param required boolean, is this argument required
	 */
	public void addArgument(String name, boolean required) {
        ArgumentEntry entry=new ArgumentEntry(name,'0',required,null);
		m_args.add(entry);

	}

	/**
	 * adds an Argument to the root of this map. 
	 * <p>A root can have multiple arguments and options
	 * @param name String,  the name for this argument
	 * @param defaultValue String, value of this argument when non is supplied.
	 */
	public void addArgument(String name, String defaultValue) {
		ArgumentEntry arg=new ArgumentEntry(name,'0',true,null);
		arg.setDefaultValue(defaultValue);
		m_args.add(arg);

	}

	/**
	 * adds an Option to the root of this map. 
	 * <p>A root can have multiple arguments and options
	 * @param longKey String , the long form of the name, used as argument name, must be unique
	 * @param shortKey char, the one character short key, must be unique
	 * @param required boolean, is the option required.
	 */
	public void addOption(String longKey, char shortKey,boolean required) {
		m_args.add(new ArgumentEntry(longKey,shortKey,required,null));
	}
	
	/**
	 * Adds an argument to an option
	 * <p>This will typically take the form of <code>-o argument</code> or <code>--option argument</code></p>
	 * @param option String, the option to add the argument to.
	 * @param name String, the name of the argument.
	 * @param required boolean, is the argument required, if the option is supplied. If the option
	 * is optional leaving the option out will not check for the required argument. However if the option
	 * is supplied it must be followed by a value for this argument.
	 */
	public void addArgumentTo(String option,String name, boolean required) {
		m_args.add(new ArgumentEntry(name,'0',required,option));
		getEntry(option).addChild(name);
	}
	
	/**
	 * adds an option to an argument
	 * <p>This means that if we defined an argument <code>argument</code>
	 * and added options <code>one, two, three</code> to it, finding any of these
	 * options will be set as value for argument. If the argument is required and none 
	 * of the defined options is supplied an ArgumentException is thrown when processing
	 * the input.
	 * @param argument
	 * @param longKey
	 * @param shortKey
	 */
	public void addOptionTo(String argument, String longKey, char shortKey) {
		m_args.add(new ArgumentEntry(longKey,shortKey,false,argument));
		getEntry(argument).addChild(longKey);
	}
	
	/**
	 * gets the Map of all arguments supplied through String[] args
	 * mapped to their argument name.
	 * @param args String[], the array of arguments typically supplied from <code> static void main(String[] args) </code>
	 * @return Map. the Map containing the argument name and argument value as defined
	 * through arguments and options. The value of the option is equivalent to its name and can be ignored.
	 * @throws ArgumentException, ArgumentException is thrown when a required argument is missing either in the 
	 * root or from an option, or when a non-defined argument or option is found when not in the root.
	 */
	public Map getList(String args[]) throws ArgumentException {
		HashMap list=new HashMap();
		processArgs(args);
		for (Iterator it=m_args.iterator();it.hasNext();) {
			ArgumentEntry entry= (ArgumentEntry) it.next();
			if (entry.getParent()==null && entry.getValue()!=null) {
				list.put(entry.getName() , entry.getValue());
			} else if (entry.getParent()==null && entry.isRequired()) {
				throw new ArgumentException(entry.getName(),ArgumentException.MISSING_ARGUMENT);
			}
				
		}
		return list;	
	}
	
	private boolean isLong(String arg) {
		if (arg.charAt(0)=='-' && arg.charAt(1)=='-') {
			return true;
		}
		return false;
	}
	
	private void processArgs(String args[]) throws ArgumentException  {
		int size=args.length;
		int idx=0;
        
		while (idx<size) {
			ArgumentEntry arg=null;
			if (isLong(args[idx])) {
				arg=findLong(args[idx]);
                if( arg!=null && processArg(arg,idx+1<size?args[idx+1]:null)) {
                    idx++;
                }
			} else if (isShort(args[idx])) {
                for (int i=1;i<args[idx].length();i++) {
                    arg=findShort(args[idx],i);
                    if (arg!=null && processArg(arg, idx+1<size?args[idx+1]:null)) {
                        idx++;
                        break;
                    }
                }
			} else {
				arg=findNextWord();
                if (arg!=null) arg.setValue(args[idx]);
			}
			if ( arg==null ) {
				throw new ArgumentException(args[idx],ArgumentException.NOT_RECOGNIZED);
			} 
			idx++;	
		}					
	}
    
    private boolean processArg(ArgumentEntry arg, String nextArg) throws ArgumentException {
        
        if (arg.getChildren()!=null) {
            ArgumentEntry child= getEntry((String) arg.getChildren().get(0));
            if (nextArg!=null && !isLong(nextArg) && !isShort(nextArg)) {
                arg.setValue(nextArg);
                return true;
            } else if (child.isRequired()) {
                throw new ArgumentException(arg.getParent(), ArgumentException.MISSING_ARGUMENT);
            }
        } else if (arg.getParent()!=null) {
            ArgumentEntry parent = getEntry(arg.getParent());
            if (parent.getValue()==null || parent.getValue()==parent.getDefaultValue()) { //Only overwrite when not set before
                 parent.setValue(arg.getName());
            }
        } else {
            arg.setValue("true");
        }
        return false;
    }
	
	private boolean isShort(String arg) {
		if (arg.charAt(0)=='-' && Character.isLetterOrDigit(arg.charAt(1))) {
			return true;
		}
		return false;
	}
	
	private ArgumentEntry findLong(String arg) {
		return getEntry(arg.substring(2));
	}
	
	private ArgumentEntry findShort(String arg, int pos) {
		char a= arg.charAt(pos);
		for (Iterator it=m_args.iterator();it.hasNext();) {
			ArgumentEntry entry= (ArgumentEntry) it.next();
			if (entry.getShortKey()==a) {
				return entry;
			}
		}
		return null;
	}

	private ArgumentEntry findNextWord() {
		for (Iterator it=m_args.iterator();it.hasNext();) {
			ArgumentEntry entry=(ArgumentEntry) it.next();
			if (entry.getShortKey()=='0' && entry.getParent()==null && entry.getChildren()==null && entry.getValue()==null) {
				return entry;
			}
		}
		return null;
	}
    
    private final ArgumentEntry getEntry(String key) {
        for (Iterator it=m_args.iterator();it.hasNext();) {
            ArgumentEntry entry=(ArgumentEntry) it.next();
            if (key.equals(entry.getName())) {
                return entry;
            }
        }
        return null;
    }
}
