/* 	DynamicClassLoader.java
 *	Copyright (c) 2006, Brains2B.org
 *	
 *	Created by: Dennis Groenendijk
 *	Created on: Feb 12, 2006
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that  * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* @#History
 * [When]		[Who]					[What]
 * Aug 8, 2008	dennis@brains2b.nl		Ignore NoClassDefFoundError when loading a class; Loading failes		
 */
package org.brains2b.loader;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

/**
 * Loads classes dynamically from a directory or jar/zip file being
 * passed in. 
 * <UL>It is not a class loader itself, but uses URLClassLoader to do
 * the actual loading. It does supply the following options:
 * <LI>find an Implementation of a Class or Interface in the classpath
 * supplied and load them, f.e. java.sql.Driver in a specific vendors driver jar</LI>
 * <LI>Load a specific class from a jar</LI>
 * <LI>Load all classes from a jar</LI>
 * </UL>
 * <p><i>Note:</i>If you load individual classes the classes referring to that
 * class are also automatically loaded as long as they are available in the classpath
 * supplied or in the bootstrap of the JVM.</p> 
 *  @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 1.00 [12-02-2006]
 */
public class DynamicClassLoader {
	public final static char FILESEP = System.getProperty("file.separator").charAt(0);
	/**
     * load a specific class identified by it's fully qualified name from the passed URLs 
     * @param name String, the fully qualified class name
     * @param url URL[], an array  of URLs in which to find the class
     * @return Class, the loaded Class. If url is null an attempt is made
     * to load the class from the current ClassLoader
     * @throws ClassNotFoundException, thrown if the class being loaded cannot be found
     * in the supplied classpath or current ClassLoader if url was null.
	 */
	public final static Class loadClass(String name, URL[] url) throws ClassNotFoundException {
		if (url==null) {
			return Class.forName(name);
		}
		URLClassLoader loader= URLClassLoader.newInstance(url);
		return loadClass(name, loader);
	}
	
	private final static Class loadClass(String name, URLClassLoader loader) throws ClassNotFoundException {
		try {
			 return Class.forName(name);
		} catch (ClassNotFoundException cnfex) {
			
		}
		Class c=loader.loadClass(name);
		if (c==null) {
			throw new ClassNotFoundException(name +" cannot be found.");
		}
		return c;
	}
    /**
     * load all the classes from the classpath supplied by the URLs
     * @param url URL[], the paths to the classes or the jar file
     * @return Class[] all the classes that have been loaded from the
     * given URLs
     * @throws ClassNotFoundException, thrown if a class cannot be loaded
     * from one of the locations supplied by the URLs.
     */
    public final static Class[] loadClasses(URL[] url) throws ClassNotFoundException {
		ArrayList l=new ArrayList();
		for (int i=0;i<url.length;i++) {
			File f=new File(url[i].getPath());
			if ( f.isDirectory() ) {
				l.addAll(loadDir(f,f));
			} else { //assume it is a zipped file
				l.addAll(loadArchive(f));
			}
		}
		Class[] classes= new Class[l.size()];
		l.toArray(classes);
		return classes;
	}
	
	/**
     * find an implementation of an <code>interface</code> or 
     * <code>class</code> in the classpath supplied by the URLs 
     * @param superOrInterface Class, the super class or interface
     * to find in the classpath
     * @param url URL[], the paths to the classes or the jar file
     * @return Class[] the classes that implement the class or interface
     * or null if it cannot be found.
     * @throws ClassNotFoundException, thrown if a class cannot be loaded
     * from one of the locations supplied by the URLs.
	 */
    public final static Class[] findImplementation(Class superOrInterface, URL[] url) throws ClassNotFoundException {
	   ArrayList l=new ArrayList();
	   Class[] all=loadClasses(url);
       for (int i=0;i<all.length;i++) {
		   if (superOrInterface.isAssignableFrom(all[i])) {
               l.add(all[i]);
           }
		}
		Class[] result=new Class[l.size()];
		l.toArray(result);
		return result;
	}
	
    /**
     * load classes from directory supplied by File.
     * @param parent File, the directory is used to truncate 
     * the file to construct the class name
     * @param f File, the file or directory referring to the class(es)
     * @return ArrayList, list of classes loaded from the dir
     * @throws ClassNotFoundException, thrown if a class cannot be loaded
     * from one of the locations supplied by the URLs.
     */
	private final static ArrayList loadDir(File parent, File f) throws ClassNotFoundException {
		ArrayList l=new ArrayList();
		File[] fs=f.listFiles();
		try {
			URLClassLoader loader =URLClassLoader.newInstance(new URL[] {f.toURL()});
    		if (fs!=null)
    			for (int i=0;i<fs.length;i++) {
    				if (fs[i].isDirectory()) {
    					l.addAll(loadDir(parent, fs[i]));
    				} else if (fs[i].isFile() && fs[i].getName().endsWith(".class")) {
    					String className=fs[i].getAbsolutePath().substring(parent.getAbsolutePath().length()+1);
    					className = className.substring(0,className.length()-6); // take of the class extention
    					className = className.replace(FILESEP,'.');
    					try {
    						Class clazz = loadClass(className,loader);
    						if (!l.contains(clazz)) {
    							l.add(clazz);
    						}
    					} catch (NoClassDefFoundError ncdef) {
    						
    					}
    				}
    			}
		} catch (MalformedURLException mfuex) {
			
		}
		return l;
	}
	
    /**
     * load classes from archive supplied by File.
     * @param f File, the name of the archive to load the classes from
     * @return ArrayList, list of classes loaded from the archive
     * @throws ClassNotFoundException, thrown if a class cannot be loaded
     * from one of the locations supplied by the URLs.
     */
	private final static ArrayList loadArchive(File f) throws ClassNotFoundException {
		ArrayList l=new ArrayList();
		try {
			URLClassLoader loader =URLClassLoader.newInstance(new URL[] {f.toURL()});
			ZipFile zf=new ZipFile(f);
			Enumeration enu= zf.entries();
			while (enu.hasMoreElements()) {
				ZipEntry e=(ZipEntry) enu.nextElement();
				if (e.getName().endsWith(".class")) {
					String className = e.getName();
					className = className.substring(0,className.length()-6); // take of the class extention
					className = className.replace('/','.');
					try {
						Class clazz = loadClass(className,loader);
						l.add(clazz);
						
					} catch (NoClassDefFoundError ncdef) {
						
					}
				}
			}
		} catch (MalformedURLException muex) {
			
		} catch (ZipException zex) {
			
		} catch (IOException iex) {
			
		}
		return l;
	}
}
